`timescale 1ns / 1ps
////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer:
//
// Create Date:    13:54:22 08/15/05
// Design Name:    
// Module Name:    packet_manager
// Project Name:   
// Target Device:  
// Tool versions:  
// Description:
//
// Dependencies:
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
////////////////////////////////////////////////////////////////////////////////
module packet_manager(clk, reset, OwnMAC, DestMAC, OwnIP, DestIP,
   startOrbit, finishOrbit, hlxId,  
	packet_send, packet_len, packet_sel,
	packet_real_sel, packet_real_read, packet_real_write, packet_real_addr, packet_real_we,packet_real_hd,
	packet_real_send, packet_real_len, 
	//packet_send_poll, packet_send_busy, packet_send_macbusy, mode_wait_send, 
	busy,
	histogram_data,histogram_addr,histogram_read_strobe,histogram_read_ack
	);																	 
    input clk, reset;
    input [47:0] OwnMAC;
    input [47:0] DestMAC;
	 input [31:0] OwnIP, DestIP;
	 input [31:0] startOrbit, finishOrbit;
	 input [7:0] hlxId;
    input packet_send;
    input [15:0] packet_real_read;
	 input [10:0] packet_len;
	 input [5:0] packet_sel;
    output [15:0] packet_real_write;
	 output [8:0] packet_real_addr;
	 output [4:0] packet_real_sel;
	 output reg packet_real_we;
	 output reg packet_real_send;
	 output reg packet_real_hd;
	 output reg [10:0] packet_real_len;
	 //input mode_wait_send;
	 output reg busy;
	 input [15:0] histogram_data;
	 output [15:0] histogram_addr;
	 input histogram_read_ack;
	 output reg histogram_read_strobe;
//	 output packet_send_poll;
	 //input packet_send_busy, packet_send_macbusy;

	reg [4:0] state = 5'h0;
	reg [15:0] header_data;
	wire [15:0] histogram_data_swapped;
	reg [8:0] packet_addr;
	reg [5:0] header_sel;
	reg [4:0] packet_fifo = 5'h0;
	reg [15:0] identCounter = 16'h0;


parameter STATE_IDLE           = 5'h0;
parameter STATE_BEGINCOPY      = 5'h1;
parameter STATE_COPY0          = 5'h2;
parameter STATE_COPY1          = 5'h3;
parameter STATE_COPY2          = 5'h4;
parameter STATE_COPY3          = 5'h5;
parameter STATE_COPY4          = 5'h6;
parameter STATE_HEADPRE        = 5'h7;
parameter STATE_HEAD0          = 5'h8;
parameter STATE_HEAD1          = 5'h9;
parameter STATE_HEAD2          = 5'ha;
parameter STATE_HEAD3          = 5'hb;
parameter STATE_HEAD4          = 5'hc;
parameter STATE_XMITSTART      = 5'hd;
parameter STATE_XMIT0          = 5'he;
parameter STATE_FINISH         = 5'hf;
parameter STATE_SUM0           = 5'h10;
parameter STATE_SUM1           = 5'h11;
parameter STATE_SUM2           = 5'h12;
parameter STATE_SUM3           = 5'h13;
parameter STATE_SUM4           = 5'h14;
parameter STATE_SUM5           = 5'h15;
parameter STATE_SUM6           = 5'h16;
parameter STATE_SUM7           = 5'h17;

parameter PacketType=16'h0800; // IP
parameter MAX_PACKET_COUNT=5'h7;
parameter UDPPort=16'h533a;
parameter HeaderBytes='d64; 

always @(posedge clk)
	busy<=(state!=STATE_IDLE);

assign histogram_data_swapped={histogram_data[7:0],histogram_data[15:8]};

assign histogram_addr[8:0]=packet_addr; // convenient!
assign histogram_addr[14:9]=header_sel;
assign histogram_addr[15]=1'h0;

	wire copy_like_mode;

assign copy_like_mode=(state==STATE_BEGINCOPY || state==STATE_COPY0 || state==STATE_COPY1 ||
								state==STATE_COPY2 ||state==STATE_COPY3 || state==STATE_COPY4);

assign packet_real_write=(copy_like_mode)?(histogram_data_swapped):(header_data);
assign packet_real_addr=packet_addr;
assign packet_real_sel=packet_fifo;


/// HEADER space
																				  
	wire [7:0] histogramSet, histogramSeq;
assign histogramSeq={5'h0,header_sel[2:0]};
assign histogramSet={5'h0,header_sel[5:3]};

	wire [11:0] payloadSize;
	wire [15:0] ipLength,udpLength, IPChecksum, UDPChecksum;

assign payloadSize=packet_real_len-HeaderBytes;
assign ipLength=packet_real_len-14;
assign udpLength=packet_real_len-14-20;

always @(posedge clk)  
  if (state==STATE_SUM0 || state==STATE_SUM1 || state==STATE_SUM2 || state==STATE_SUM3) header_data<=IPChecksum;
  else if (state==STATE_SUM4 || state==STATE_SUM5|| state==STATE_SUM6 || state==STATE_SUM7) header_data<=UDPChecksum;
  else begin
   case (packet_addr[4:0]) 
	5'h00 : header_data<=DestMAC[47:32];
	5'h01 : header_data<=DestMAC[31:16];
	5'h02 : header_data<=DestMAC[15:0];
	5'h03 : header_data<=OwnMAC[47:32];
	5'h04 : header_data<=OwnMAC[31:16];
	5'h05 : header_data<=OwnMAC[15:0];
	5'h06 : header_data<=PacketType;		
	5'h07 : header_data<=16'h4500; // {Service,Len,Vers}
	5'h08 : header_data<=ipLength; // total length	- eth
	5'h09 : header_data<=identCounter; // ident counter
	5'h0a : header_data<=16'h4000; // do not fragment
	5'h0b : header_data<=16'h0311; // {ttl, protocol}
	5'h0c : header_data<=16'h0; // {checksum, come back to}
	5'h0d : header_data<=OwnIP[31:16]; // SRC IP
	5'h0e : header_data<=OwnIP[15:0]; // SRC IP
	5'h0f : header_data<=DestIP[31:16]; // DEST IP
	5'h10 : header_data<=DestIP[15:0]; // DEST IP
	5'h11 : header_data<=UDPPort; // SRC PORT
	5'h12 : header_data<=UDPPort; // DEST PORT
   5'h13 : header_data<=udpLength; // total - eth - IP
	5'h14 : header_data<=udpLength; // {checksum, come back to (this is a trick!)}

/// FROM HERE DOWN, WE BYTE-SWAP for Linux
	5'h15 : header_data<={4'h2,4'h1,4'h0,4'h0}; // Retransmit,Subtype,Type,Version
	5'h16 : header_data<={hlxId,8'h0}; 
	5'h17 : header_data<={startOrbit[7:0],startOrbit[15:8]};
	5'h18 : header_data<={startOrbit[23:16],startOrbit[31:24]};
	5'h19 : header_data<={finishOrbit[7:0],finishOrbit[15:8]};
	5'h1a : header_data<={finishOrbit[23:16],finishOrbit[31:24]};
	5'h1b : header_data<={histogramSet,histogramSeq};
	5'h1c : header_data<={payloadSize[7:0],packet_fifo[3:0],payloadSize[11:8]};
	5'h1d : header_data<={16'h0};
	5'h1e : header_data<={16'h0};
	5'h1f : header_data<={16'h0};
	endcase
end


   wire [8:0] inc_addr;
assign inc_addr=packet_addr+1;
	reg [2:0] xmit_timer;

// packet manager state machine

always @(posedge clk)
	if (reset) begin
		state<=STATE_IDLE;
		packet_real_we<=1'h0;
		histogram_read_strobe<=1'h0;
		packet_fifo<=0;
		identCounter<=16'h0;
	end
	else begin
		case (state) 
			STATE_IDLE : if (packet_send) begin
								state<=STATE_BEGINCOPY;
								header_sel<=packet_sel;
								packet_real_len<=packet_len+HeaderBytes; // header added on
							 end else state<=STATE_IDLE;
		   STATE_BEGINCOPY : begin
										packet_addr<=9'h0;
										state<=STATE_COPY0;
										packet_real_hd<=1'h0;
										identCounter<=identCounter+1;
									end
			STATE_COPY0 : begin
									histogram_read_strobe<=1'h1;
									state<=STATE_COPY1;
							  end
			STATE_COPY1 : if (histogram_read_ack) begin 
									state<=STATE_COPY2;
									packet_real_we<=1'h1;
							  end else state<=STATE_COPY1;
			STATE_COPY2 :  state<=STATE_COPY3;
			STATE_COPY3 : begin
									packet_real_we<=1'h0;
									histogram_read_strobe<=1'h0;
									if (!histogram_read_ack) state<=STATE_COPY4;
									else state<=STATE_COPY3;
							  end
			STATE_COPY4 : begin
								 if (payloadSize[9:1]==inc_addr) state<=STATE_HEADPRE; // end of copy
								 else state<=STATE_COPY0; // continue copy
				 				 packet_addr[8:0]<=inc_addr;
							  end
			STATE_HEADPRE : begin
									packet_addr<=9'h0;
									state<=STATE_HEAD0;
									packet_real_hd<=1'h1;
								 end
			STATE_HEAD0 : state<=STATE_HEAD1;
			STATE_HEAD1 : begin
								 state<=STATE_HEAD2;
								 packet_real_we<=1'h1;
							  end
			STATE_HEAD2	: state<=STATE_HEAD3;
			STATE_HEAD3	: begin 
								 packet_real_we<=1'h0;
								 state<=STATE_HEAD4;
								end
			STATE_HEAD4 : begin
								if (inc_addr==9'h20) state<=STATE_SUM0;
								else begin
									packet_addr<=inc_addr;
									state<=STATE_HEAD0;
								end
							  end
         STATE_SUM0 : begin				   // IP Checksum
			               packet_addr<=9'h0c;
								state<=STATE_SUM1;
							 end
         STATE_SUM1 : begin
			               packet_real_we<=1'h1;
								state<=STATE_SUM2;
							 end
			STATE_SUM2 : state<=STATE_SUM3;
         STATE_SUM3 : begin
								packet_real_we<=1'h0;
								state<=STATE_SUM4;														  
							 end
         STATE_SUM4 : begin					// UDP Checksum
			               packet_addr<=9'h14;
								state<=STATE_SUM5;
							 end
         STATE_SUM5 : begin
			               packet_real_we<=1'h1;
								state<=STATE_SUM6;
							 end
			STATE_SUM6 : state<=STATE_SUM7;
         STATE_SUM7 : begin
								packet_real_we<=1'h0;
								state<=STATE_XMITSTART;														  
							 end
		  STATE_XMITSTART: begin
		  				xmit_timer<=3'h0;
						state<=STATE_XMIT0;
					end
		  STATE_XMIT0: begin
		  				xmit_timer<=xmit_timer+1;
						packet_real_send<=(xmit_timer<3'h4);
						if (xmit_timer==3'h7) state<=STATE_FINISH;
						else state<=STATE_XMIT0;
					end

		  STATE_FINISH : begin
		  				if (packet_fifo==MAX_PACKET_COUNT) packet_fifo<=0;
						else packet_fifo<=packet_fifo+1;
						state<=STATE_IDLE;
					end
		endcase
	end


	wire doIPAdd, doUDPAdd;
	wire [15:0] udp_data;
ipChecksum ipSum(.sysclk(clk),.reset(state==STATE_IDLE),.data(header_data),.add(doIPAdd),.checksum(IPChecksum));
ipChecksum udpSum(.sysclk(clk),.reset(state==STATE_IDLE),.data(udp_data),.add(doUDPAdd),.checksum(UDPChecksum));

assign doIPAdd = (state==STATE_HEAD3 && packet_addr[4:0]>5'h6 && packet_addr[4:0]<5'h11);
assign doUDPAdd = (state==STATE_HEAD3 && packet_addr[4:0]>5'hb) || (state==STATE_COPY3);

assign udp_data = (copy_like_mode)?(histogram_data_swapped):
						(packet_addr[4:0]==5'hc)?(16'h0011): // Meta-header piece
						(header_data);

endmodule
